/* $Id: pbjtime.c,v 1.57 2003/06/07 06:51:04 chrisf Exp $ */

/*
PBJTime - Clock utility
Copyright (C) 2003 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <PalmOS.h>
#include <Handspring/HsExt.h>
#include "pbjtimeRsc.h"
#include "pbjtime.h"

/* globals */
UInt16 tps;
Boolean ispopup=false;
DmOpenRef MainDB;
Boolean is_handspring;

/* prefs */
struct sprefs prefs;

Boolean is_jog(EventPtr e)
{
    if (!is_handspring) return false;
    return (HsEvtMetaEvent(e) == hsEvtPageKeyIsJog);
}

Boolean MainFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    static Int32 oldticks=0;
    Int32 ticks;
    UInt16 i;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        DrawGadgets(frm, true);
        SetFieldHandles(frm, MainDB);
        handled = true;
        break;

    case frmUpdateEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        DrawGadgets(frm, true);
        handled = true;
        break;

    case frmCloseEvent:
        frm = FrmGetActiveForm();
        /* stop referencing our record */
        ClearFieldHandles(frm);
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        case mitemID_about:
            FrmPopupForm(AboutForm);
            break;
        case mitemID_prefs:
            FrmPopupForm(PrefsForm);
            break;
        case mitemID_clearbuffer:
            frm = FrmInitForm(ClearBufferForm);
            i = FrmDoDialog(frm);
            FrmDeleteForm(frm);
            if (i == buttonID_clearbufferyes) {
                frm = FrmGetActiveForm();
                ClearFieldHandles(frm);
                /* remove rec #0 */
                if (DmQueryRecord(MainDB, 0)) DmRemoveRecord(MainDB, 0);
                SetFieldHandles(frm, MainDB);
            }
            break;
        }
       handled = true;
       break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        }
        break;

    case popSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        }
        break;

    case keyDownEvent:
        frm = FrmGetActiveForm();
        switch (e->data.keyDown.chr) {
        case pageUpChr:
            TextFieldPageScroll(frm, winUp, is_jog(e));
            break;
        case pageDownChr:
            TextFieldPageScroll(frm, winDown, is_jog(e));
            break;
        }
        break;

    case sclRepeatEvent:
        frm = FrmGetActiveForm();
        switch (e->data.sclRepeat.scrollBarID) {
        case scrollbarID_text:
            TextFieldScroll(frm, e->data.sclRepeat.newValue-
                e->data.sclRepeat.value);
            break;
        }
        break;

    case nilEvent:
        ticks = TimGetTicks();
        /* every second update the values */
        if ((ticks-oldticks) >= (tps*10)) {
            oldticks = ticks;
            frm = FrmGetActiveForm();
            DrawGadgets(frm, false);
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean PopupFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    static Int32 oldticks=0;
    Int32 ticks;

    handled = false;

    /* if user clicks outside my popup it may change the draw window */
    frm = FrmGetActiveForm();
    WinSetDrawWindow(FrmGetWindowHandle(frm));

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        DrawGadgets(frm, true);
        handled = true;
        break;

    case frmUpdateEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        DrawGadgets(frm, true);
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        }
        break;

    case popSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        }
        break;

    case keyDownEvent:
        frm = FrmGetActiveForm();
        switch (e->data.keyDown.chr) {
        case pageUpChr:
            TextFieldPageScroll(frm, winUp, is_jog(e));
            break;
        case pageDownChr:
            TextFieldPageScroll(frm, winDown, is_jog(e));
            break;
        }
        break;

    case sclRepeatEvent:
        frm = FrmGetActiveForm();
        switch (e->data.sclRepeat.scrollBarID) {
        case scrollbarID_text:
            TextFieldScroll(frm, e->data.sclRepeat.newValue-
                e->data.sclRepeat.value);
            break;
        }
        break;

    case nilEvent:
        ticks = TimGetTicks();
        /* every second update the values */
        if ((ticks-oldticks) >= (tps*10)) {
            oldticks = ticks;
            frm = FrmGetActiveForm();
            DrawGadgets(frm, false);
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean PrefsFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    FieldPtr fld;
    char *txt, txt2[6];

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);

        /* duration */
        fld = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, fieldID_prefspopupduration));
        FldDelete(fld, 0, FldGetMaxChars(fld));
        StrPrintF(txt2, "%d", prefs.popupduration);
        FldInsert(fld, txt2, StrLen(txt2));
        FldSetDirty(fld, false); /* so I know when the user changes it */

        handled = true;
        break;

    case frmCloseEvent:
        frm = FrmGetActiveForm();
        /* need to convert duration text into UInt16 */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_prefspopupduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.popupduration = (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
        }
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        frm = FrmGetActiveForm();
        /* need to convert duration text into UInt16 */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_prefspopupduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.popupduration = (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
        }
        switch (e->data.ctlSelect.controlID) {
        case buttonID_prefsok:
            FrmReturnToForm(0);
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean AboutFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_aboutok:
            FrmReturnToForm(0);
            handled = true;
            break;
        case buttonID_aboutcontrib:
            FrmPopupForm(ContribForm);
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean ContribFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    ListPtr lst;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_contrib));
        /* turn off the highlight */
        LstSetSelection(lst, -1);
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_contribok:
            FrmReturnToForm(0);
            handled = true;
            break;
        }
        break;

    case keyDownEvent:
        if (e->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, winUp, LstGetVisibleItems(lst));
        } else if (e->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, winDown, LstGetVisibleItems(lst));
        }
        break;

    default:
        break;
    }

    return handled;
}

void DrawGadgets(FormPtr frm, Boolean forcedraw)
{
    UInt16 battv;
    UInt8 battp;
    static UInt16 obattv;
    static UInt8 obattp;
    char txt[64];
    RectangleType r;
    UInt32 tgs;
    UInt16 i, l;
    IndexedColorType oldict;

    if ((frm != FrmGetFormPtr(MainForm)) &&
        (frm != FrmGetFormPtr(PopupForm))) return;
    if (FrmGetWindowHandle(frm) != WinGetDrawWindow()) return;

    WinPushDrawState();
    oldict =
        WinSetBackColor(UIColorGetTableEntryIndex(UILastColorTableEntry+1));

    battv = SysBatteryInfo(false, NULL, NULL, NULL, NULL, NULL, &battp);
    if (forcedraw || (battv != obattv) || (battp != obattp)) {
        StrPrintF(txt, "%d.%02dv %d%%", (int)(battv/100), (int)(battv%100),
            (int)battp);
        l = StrLen(txt);
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_batt), &r);
        WinEraseRectangle(&r, 0);
        i = (r.extent.x-FntCharsWidth(txt, l))>>1;
        WinDrawChars(txt, l, r.topLeft.x+i, r.topLeft.y);
        WinDrawRectangleFrame(simpleFrame, &r);
        obattv = battv;
        obattp = battp;
    }

    tgs = TimGetSeconds();
    l = GetDateText(tgs, txt);
    FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_date), &r);
    WinEraseRectangle(&r, 0);
    i = (r.extent.x-FntCharsWidth(txt, l))>>1;
    WinDrawChars(txt, l, r.topLeft.x+i, r.topLeft.y);
    WinDrawRectangleFrame(simpleFrame, &r);

    l = GetTimeText(tgs, txt);
    FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_time), &r);
    WinEraseRectangle(&r, 0);
    i = (r.extent.x-FntCharsWidth(txt, l))>>1;
    WinDrawChars(txt, l, r.topLeft.x+i, r.topLeft.y);
    WinDrawRectangleFrame(simpleFrame, &r);

    WinSetBackColor(oldict);
    WinPopDrawState();
}

UInt16 GetDateText(UInt32 sec, char *txt)
{
    DateType dt;
    DateSecondsToDate(sec, &dt);
    DateToDOWDMFormat(dt.month, dt.day, dt.year+1904, dfMDYWithSlashes, txt);
    return StrLen(txt);
}

UInt16 GetTimeText(UInt32 sec, char *txt)
{
    DateTimeType dtt;
    TimSecondsToDateTime(sec, &dtt);
    TimeToAscii(dtt.hour, dtt.minute, tfColonAMPM, txt);
    return StrLen(txt);
}

void SetFieldHandles(FormPtr frm, DmOpenRef db)
{
    FieldPtr fld;
    MemHandle rech;

    if ((frm != FrmGetFormPtr(MainForm)) &&
        (frm != FrmGetFormPtr(PopupForm))) return;
    if (FrmGetWindowHandle(frm) != WinGetDrawWindow()) return;
    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_text));
    if ((rech = DmQueryRecord(db, 0 /* rec #0 */))) {
        FldSetTextHandle(fld, rech);
    } else {
        FldSetTextHandle(fld, NULL);
    }
    FldRecalculateField(fld, false);
    TextUpdateScrollBar(frm, true);
    FldDrawField(fld);
}

void ClearFieldHandles(FormPtr frm)
{
    FieldPtr fld;

    if ((frm != FrmGetFormPtr(MainForm)) &&
        (frm != FrmGetFormPtr(PopupForm))) return;
    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_text));
    FldSetTextHandle(fld, NULL);
}

void TextUpdateScrollBar(FormPtr frm, Boolean init)
{
    UInt16 scrollPos;
    UInt16 textHeight;
    UInt16 fieldHeight;
    Int16 maxValue;
    FieldPtr fld;
    ScrollBarPtr bar;

    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_text));
    bar = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, scrollbarID_text));
    FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);
    if (textHeight > fieldHeight) {
        maxValue = textHeight-fieldHeight;
    } else if (scrollPos) {
        maxValue = scrollPos;
    } else {
        maxValue = 0;
    }
    if (init) {
        /* init the field and scrollbar to bottom */
        scrollPos = maxValue;
    }
    SclSetScrollBar(bar, scrollPos, 0, maxValue, fieldHeight-1);
    FldScrollField(fld, scrollPos, winDown);
}

void TextFieldScroll(FormPtr frm, Int16 linesToScroll)
{
    UInt16 blankLines;
    Int16 min;
    Int16 max;
    Int16 value;
    Int16 pageSize;
    FieldPtr fld;
    ScrollBarPtr bar;

    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_text));
    bar = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, scrollbarID_text));

    if (linesToScroll < 0) {
        blankLines = FldGetNumberOfBlankLines(fld);
        FldScrollField(fld, -linesToScroll, winUp);
        /*
         * If there were blank lines visible at the end of the field
         * then we need to update the scroll bar.
         */
        if (blankLines) {
            /* Update the scroll bar */
            SclGetScrollBar(bar, &value, &min, &max, &pageSize);
            if (blankLines > -linesToScroll) max += linesToScroll;
            else max -= blankLines;
            SclSetScrollBar(bar, value, min, max, pageSize);
        }
    } else if (linesToScroll > 0) FldScrollField (fld, linesToScroll, winDown);
}

void TextFieldPageScroll(FormPtr frm, WinDirectionType direction,
    Boolean onlyoneline)
{
    Int16 min;
    Int16 max;
    Int16 value;
    Int16 pageSize;
    UInt16 linesToScroll;
    FieldPtr fld;
    ScrollBarPtr bar;

    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_text));
    bar = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, scrollbarID_text));

    if (FldScrollable(fld, direction)) {
        if (onlyoneline) linesToScroll = 1;
        else linesToScroll = FldGetVisibleLines(fld)-1;
        FldScrollField(fld, linesToScroll, direction);
        /* Update the scroll bar */
        SclGetScrollBar(bar, &value, &min, &max, &pageSize);
        if (direction == winUp) value -= linesToScroll;
        else value += linesToScroll;
        SclSetScrollBar(bar, value, min, max, pageSize);
    }
}

/*
 * Returns field that has the focus, if any, including in embedded tables.
 */
FieldPtr GetFocusObjectPtr(void)
{
    FormPtr frm;
    UInt16 focus;
    FormObjectKind objType;

    frm = FrmGetActiveForm();
    focus = FrmGetFocus(frm);
    if (focus == noFocus) return NULL;
    objType = FrmGetObjectType(frm, focus);
    if (objType == frmFieldObj) return FrmGetObjectPtr(frm, focus);
    else if (objType == frmTableObj)
        return TblGetCurrentField(FrmGetObjectPtr(frm, focus));
    return NULL;
}

Boolean FormHandleMenuEvent(UInt16 menuID)
{
    FieldPtr fld;

    switch (menuID) {
    case sysEditMenuUndoCmd:
    case sysEditMenuCutCmd:
    case sysEditMenuCopyCmd:
    case sysEditMenuPasteCmd:
    case sysEditMenuSelectAllCmd:
        fld = GetFocusObjectPtr();
        if (!fld) return false;
        if (menuID == sysEditMenuUndoCmd) FldUndo(fld);
        else if (menuID == sysEditMenuCutCmd) FldCut(fld);
        else if (menuID == sysEditMenuCopyCmd) FldCopy(fld);
        else if (menuID == sysEditMenuPasteCmd) FldPaste(fld);
        else if (menuID == sysEditMenuSelectAllCmd)
            FldSetSelection (fld, 0, FldGetTextLength (fld));
        return true;
    case sysEditMenuKeyboardCmd:
        SysKeyboardDialog(kbdDefault);
        return true;
    case sysEditMenuGraffitiCmd:
        SysGraffitiReferenceDialog(referenceDefault);
        return true;
    }
    return false;
}

Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    UInt16 formId;
    Boolean handled;

    handled = false;
    if (e->eType == frmLoadEvent) {
        formId = e->data.frmLoad.formID;
        frm = FrmInitForm(formId);
        FrmSetActiveForm(frm);

        switch(formId) {
        case MainForm:
            FrmSetEventHandler(frm, MainFormHandleEvent);
            break;
        case PrefsForm:
            FrmSetEventHandler(frm, PrefsFormHandleEvent);
            break;
        case AboutForm:
            FrmSetEventHandler(frm, AboutFormHandleEvent);
            break;
        case ContribForm:
            FrmSetEventHandler(frm, ContribFormHandleEvent);
            break;
        }
        handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
UInt16 StartApplication(void)
{
    UInt32 tui32;

    tps = SysTicksPerSecond();
    LoadPrefs(&prefs);

    /* determine if we have handspring extensions */
    if (!FtrGet(hsFtrCreator, hsFtrIDVersion, &tui32)) is_handspring = true;
    if (!OpenDatabase(&MainDB)) return 1;
    /* this feature is available whenever we have a good MainDB */
    FtrSet(MainAppID, FTR_MAINDB, (UInt32)MainDB);

    UIColorPushTable();
    UIColorSetTableEntry(UIFormFill, &(RGBColorType){0, 0x00, 0xff, 0xff});
    UIColorSetTableEntry(UIFieldBackground,
        &(RGBColorType){0, 0x00, 0xff, 0xff});
//    UIColorSetTableEntry(UIFormFill, &(RGBColorType){0, 0xc8, 0xcc, 0xf8});
//    UIColorSetTableEntry(UIFieldBackground,
//        &(RGBColorType){0, 0xc8, 0xcc, 0xf8});
    UIColorSetTableEntry(UILastColorTableEntry+1,
        &(RGBColorType){0, 0xff, 0xff, 0x00});

    if (!ispopup) FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
void StopApplication(void)
{
    UIColorPopTable();
    if (!ispopup) {
        FrmSaveAllForms();
        FrmCloseAllForms();
    }
    CloseDatabase(MainDB);
    /* this feature is available whenever we have a good MainDB */
    FtrUnregister(MainAppID, FTR_MAINDB);
}

/* The main event loop */
void EventLoop(void)
{
    Err err;
    EventType e;
    Int32 eventTimeout;
    Boolean stoploop=false;
    Boolean ishelddown=true;
    Boolean tonestopped;
    UInt32 stoptime=0;
    UInt32 tui32;

    eventTimeout = tps/4;

    do {

    /* need to know that hotkey has been released */
    if (ispopup && !stoploop && !KeyCurrentState()) {
        ishelddown = false;
        stoploop = true; /* testing */
        stoptime = TimGetSeconds()+prefs.popupduration;
    }

    EvtGetEvent(&e, eventTimeout);

    /* this feature may be changed by other launchcode invocations */
    if (!FtrGet(MainAppID, FTR_TONESTOPPED, &tui32) && (tui32 == 1)) {
        tonestopped = true;
    } else {
        tonestopped = false;
    }

    if (ispopup && (e.eType == keyDownEvent)) {
        if (!(e.data.keyDown.modifiers&poweredOnKeyMask)) {
            switch (e.data.keyDown.chr) {
            case hard1Chr:
            case hard2Chr:
            case hard3Chr:
            case hard4Chr:
            case hsChrJogScan:
            case chrLineFeed:
                /* ignore autorepeat of hardchrs */
                if (e.data.keyDown.modifiers&autoRepeatKeyMask) continue;
                /* must have released hotkey at least once */
                stoploop = !ishelddown;
                if (ishelddown) continue;
                stoptime = 0;
                /* no break wanted */
            case pageUpChr:
            case pageDownChr:
            case vchrPowerOff:
                /* user activity stops the playing tone */
                if (!tonestopped) {
                    treohelper_stop_tone();
                    FtrSet(MainAppID, FTR_TONESTOPPED, 1);
                }
                break;
            }
        } else {
        }
    }

    /* user activity stops the playing tone */
    if (!tonestopped && (e.eType == penDownEvent)) {
        treohelper_stop_tone();
        FtrSet(MainAppID, FTR_TONESTOPPED, 1);
    }

    if (stoploop && (TimGetSeconds() >= stoptime)) break;

    /* any event other than nil gives more delay; i.e. key,tap, EV_KEEPALIVE */
    if (stoploop && (e.eType != nilEvent))
        stoptime = TimGetSeconds()+prefs.popupduration;

    /* exit popup if auto-off */
    if (ispopup && (e.eType == keyDownEvent) &&
        ((e.data.keyDown.modifiers & commandKeyMask) &&
        (e.data.keyDown.chr == vchrPowerOff))) {
        EvtAddEventToQueue(&e); /* add it back into the queue */
        break; /* exit our event loop to get rid of the popup */
    }

    if (!SysHandleEvent(&e)) {
        if (!MenuHandleEvent(NULL, &e, &err)) {
            if (!ApplicationHandleEvent(&e)) {
                FrmDispatchEvent(&e);
            }
        }
    }

//    if (ispopup && !KeyCurrentState()) break;

    } while ((e.eType != appStopEvent) && (e.eType != EV_STOPLOOP));

}

void LoadPrefs(struct sprefs *prefs)
{
    Int16 prefsVersion;
    UInt16 prefsSize;

    prefs->popupduration = 30;
    prefsSize = sizeof(prefs->popupduration);
    prefsVersion = PrefGetAppPreferences(MainAppID, 0 /* Pref ID */,
        &prefs->popupduration, &prefsSize, true);
}

void SavePrefs(struct sprefs *prefs)
{
    PrefSetAppPreferences(MainAppID, 0 /* Pref ID */, 1 /* Version */,
        &prefs->popupduration, sizeof(prefs->popupduration), true);
}

void post_custom_event(UInt16 etype)
{
    EventType new_event;
    new_event.eType = etype;
    EvtAddEventToQueue(&new_event);
}

Boolean OpenDatabase(DmOpenRef *db)
{
    UInt16 cardNo;
    LocalID dbID;
    UInt16 attributes;

    /* Set these to zero in case any of the opens fail */
    *db = 0;

    /* Open our database */
    *db = DmOpenDatabaseByTypeCreator(MainDBType, MainAppID,
        dmModeReadWrite);

    if (!*db) {
        /* Create our database if it doesn't exist yet */
        if (DmCreateDatabase(0, MainDBName, MainAppID, MainDBType, false))
            return false;
        *db = DmOpenDatabaseByTypeCreator(MainDBType, MainAppID,
            dmModeReadWrite);
        if (!*db) return false;
        /* Set the backup bit.  This is to aid syncs with non Palm software. */
        DmOpenDatabaseInfo(*db, &dbID, NULL, NULL, &cardNo, NULL);
        DmDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL);
        attributes |= dmHdrAttrBackup;
        DmSetDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }

    return true;
}

void CloseDatabase(DmOpenRef db)
{
    if (db) {
        DmCloseDatabase(db);
    }
}

void SetTimeOfNextAlarm(UInt32 alarmTime, UInt32 ref, AlmAlarmProcPtr procP)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (procP) {
        AlmSetProcAlarm(procP, ref, alarmTime);
    } else {
        if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
            sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
        AlmSetAlarm(cardNo, dbID, ref, alarmTime, 0);
    }
}

void DisplayAlarm(UInt16 launchFlags)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 res;
    FormActiveStateType formState;

    /* launch the popup */
    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, MainAppID, true, &cardNo, &dbID) == 0)) {
        /* save all of the current app's UI state */
        FrmSaveActiveState(&formState);
        SysAppLaunch(cardNo, dbID,
            sysAppLaunchFlagNewStack|sysAppLaunchFlagNewGlobals,
            sysAppLaunchCmdCustomBase+1, NULL, &res);
        /* restore all of the current app's UI state */
        FrmRestoreActiveState(&formState);
    }
}

void treohelper_stop_tone(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 res;
    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, 'THEL', true, &cardNo, &dbID) == 0)) {
        SysAppLaunch(cardNo, dbID, 0, sysAppLaunchCmdCustomBase+1,
            NULL, &res);
    }
}

/*
 * I put these here because I didn't want my PilotMain() to use up all that
 * stack space for each and every launchcode out there.
 */
UInt32 HandleLaunchCodes(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
    Err err;
    FormPtr frm;
    FontID fnt;
    RectangleType r;
    Int16 l, l2, l3, l4, over;
    DmOpenRef db;
    MemHandle rech;
    UInt16 reci;
    void *p;
    Boolean isglobaldb;
    UInt32 tui32;
    char dts[64], *dtsp=dts;
    char dt[16], tm[16];
    UInt16 dtsl;
    UInt32 tgs;
    UInt32 retval=0;
    const char dtsf[]=
        "\n\227\227\227 %s %s \227\227\227\n";
    struct sprefs tprefs;

    LoadPrefs(&tprefs);

    switch (cmd) {

    case sysAppLaunchCmdCustomBase+1:
        /* currently in application? */
        if (!(launchFlags&sysAppLaunchFlagNewGlobals)) {
            return 0;
        }

        /* set an idle power off situation */
        EvtSetAutoOffTimer(SetExactly, tprefs.popupduration);

        ispopup = true;
        err = StartApplication();
        if (err) return err;
        fnt = FntSetFont(stdFont);
        frm = FrmInitForm(PopupForm);
        FrmSetActiveForm(frm);
        FrmSetEventHandler(frm, PopupFormHandleEvent);
        FrmDrawForm(frm);
        FrmGetFormBounds(frm, &r);
        r.topLeft.x = 1;
        r.topLeft.y = 1;
        r.extent.x -= 3;
        r.extent.y -= 3;
        WinDrawRectangleFrame(popupFrame, &r);
        DrawGadgets(frm, true);
        SetFieldHandles(frm, MainDB);
        EventLoop();
        ClearFieldHandles(frm);
        StopApplication();
        FrmEraseForm(frm);
        FrmDeleteForm(frm);
        FntSetFont(fnt);
        FrmSetActiveForm(NULL);
        break;

    case sysAppLaunchCmdCustomBase+100:
        /* launched from TreoHelper */
        if (!cmdPBP) break;

        /* so I can turn it off with user activity */
        FtrSet(MainAppID, FTR_TONESTOPPED, 0);

        /* currently in app or popup?  then it needs to use shared DmOpenRef */
        if (!FtrGet(MainAppID, FTR_MAINDB, &tui32)) {
            db = (DmOpenRef)tui32;
            isglobaldb = true;
            /* temporarily take the rechandle from this field */
            frm = FrmGetActiveForm();
            ClearFieldHandles(frm);
        } else {
            if (!OpenDatabase(&db)) break;
            isglobaldb = false;
        }

        /* build date/time stamp */
        tgs = TimGetSeconds();
        GetDateText(tgs, dt);
        GetTimeText(tgs, tm);
        StrPrintF(dts, dtsf, dt, tm);
        dtsl = StrLen(dts);

        /* save off the SMS text into our database */
        l = StrLen(cmdPBP)+1;
        do {
        reci = 0;
        if ((rech = DmQueryRecord(db, reci /* rec #0 */)) == 0) {
            /* don't include the first \n */
            dtsp++;
            dtsl--;
            /* create new rec #0 */
            rech = DmNewRecord(db, &reci, l+dtsl);
            l2 = 0;
        } else {
            /* get orig size */
            l3 = MemHandleSize(rech);
            l2 = (l3 >= 1) ? l3-1 : 0; /* get rid of NULL */
            /* l4 is new size */
            l4 = l2+l+dtsl;
            /* but we want to cap it at MAX_TEXT_SIZE */
            over = l4-MAX_TEXT_SIZE;
            if (over > 0) { /* over our cap */
                /* scroll the rec so that our new size will equal cap */
                p = MemHandleLock(rech);
                DmWrite(p, 0, &((UInt8 *)p)[over], l3-over);
                MemPtrUnlock(p);
                l2 -= over; /* adjust for shrinkage */
                l4 -= over;
            }
            /* resize to include new SMS */
            if (!(rech = DmResizeRecord(db, reci /* rec #0 */, l4)))
                rech = DmGetRecord(db, reci /* rec #0 */);
        }
        if (!rech) break;
        p = MemHandleLock(rech);
        /* need to write delim if it isn't the first one */
        DmWrite(p, l2, dtsp, dtsl);
        l2 += dtsl;
        /* tack new SMS onto end, NULL terminated */
        DmWrite(p, l2, cmdPBP, l);
        MemPtrUnlock(p);
        DmReleaseRecord(db, reci, true);
        } while (0);

        if (isglobaldb) {
            /* assign the rechandle back to this field */
            frm = FrmGetActiveForm();
            SetFieldHandles(frm, db);
            post_custom_event(EV_KEEPALIVE); /* keep popup showing */
            /* don't want to launch it since we're already in it */
        } else {
            CloseDatabase(db);
            /* set a timer to launch our popup */
            SetTimeOfNextAlarm(TimGetSeconds(), 0, NULL);
        }

        break;

    default:
        retval = sysErrParamErr;
    }
    return retval;
}

/*
 * The program starts here.
 */
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
    Err err;
    UInt32 romVersion;
    UInt32 theInstalledRAM; 

    switch (cmd) {
    case sysAppLaunchCmdNormalLaunch:
        /*
         * This ROM version test comes from Palm's Address application.
         * Thankfully too because I would hate to guess at this kind of
         * stuff and make some poor guy's Pilot blow up.
         */
        FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
        /*
         * Need at least a PalmPilot Pro because of the TCP/IP.
         */
        if (romVersion < 0x03000000) {
            if ((launchFlags &
                (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
                (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
                /*
                 * Pilot 1.0 will continuously relaunch this app unless we
                 * switch to another safe one.
                 */
                if (romVersion < 0x02000000) {
                    FrmAlert(RomIncompatibleAlert);
                    AppLaunchWithCommand(sysFileCDefaultApp,
                        sysAppLaunchCmdNormalLaunch, NULL);
                }
                err = MemCardInfo( 0, NULL, NULL, NULL, NULL, NULL,
                    &theInstalledRAM, NULL); 
                if (!err && (theInstalledRAM == 524288)) { 
                    /*
                     * We are running on a PalmPilot Personal unit
                     */
                    FrmAlert(RomIncompatibleAlert);
                    return sysErrRomIncompatible;
                }
            }
        }
        /*
         * I guess we are an appropriate ROM version so let's run.
         */
        err = StartApplication();
        if (err) return err;
        EventLoop();
        StopApplication();
        break;

    case sysAppLaunchCmdDisplayAlarm:
        DisplayAlarm(launchFlags);
        break;

    case sysAppLaunchCmdCustomBase+1:
    case sysAppLaunchCmdCustomBase+100:
        return HandleLaunchCodes(cmd, cmdPBP, launchFlags);
        break;

    default:
        return sysErrParamErr;
    }

    return 0;
}
